@@ -1,5 +1,5 @@
 import pandas as pd
-import json
+import json, re
 from gql import gql
 from gql import Client as gql_Client
 from gql.transport.requests import RequestsHTTPTransport
@@ -7,16 +7,12 @@ from .types import __VALUE_TYPES__, DATASET, INNER_DATASET, __PATTERNS__, WM__PA
 from requests.exceptions import *  
 from datetime import datetime, timedelta
 from time import sleep
-### DEBUGGER
-# import pprint
-# printy =  pprint.PrettyPrinter(indent=4).pprint
-##
 
 class safeGraphError(Exception):
     pass
 
 class HTTP_Client:
-    def __init__(self, apikey, max_tries=10):
+    def __init__(self, apikey, max_tries=3):
         self.df = pd.DataFrame()
         self.lst = [] 
         self.url = 'https://api.safegraph.com/v1/graphql'
@@ -76,9 +72,7 @@ class HTTP_Client:
             for i in range(1, week_loop+1):
                 self._date.append((start_of_week1 + timedelta(days=7*i)).strftime("%Y-%m-%d"))
         else:
-            # XXX
-            # better explain
-            raise safeGraphError(f'''*** Unrecognized DateTime! {value}\n must be either list/str/dict''')
+            raise safeGraphError(f'''*** Unrecognized DateTime! {value}\n must be either one list/str/dict''')
 
     def __change_value_types_pandas(self):
         for key, val in __VALUE_TYPES__.items():
@@ -133,10 +127,24 @@ class HTTP_Client:
         w_run_ = 1 # for weekly patterns
         query = ""
         data_type = []
+        try:
+            cols = columns.copy()
+        except AttributeError:
+            cols = columns # cols is *
         if product == "safegraph_core.*":
             # if all data from safegraph_core
             pattern_pick = __PATTERNS__["safegraph_core"]
-            available_columns = (lambda x: [j for j in pattern_pick if j not in ["__header__", "__footer__"]] if x=="*" else [j for j in pattern_pick if j in columns] )(columns)#[j for j in pattern_pick if j in columns]
+            available_columns = (lambda x: [j for j in pattern_pick if j not in ["__header__", "__footer__"]] if x=="*" else [j for j in pattern_pick if j in cols] )(cols)#[j for j in pattern_pick if j in columns]
+            if type(cols) != str:
+                # XXX
+                # TEST
+                try:
+                    cols.remove("placekey")
+                except ValueError:
+                    pass
+                if len(available_columns) < len(cols):
+                    errors = [i for i in cols if i not in pattern_pick]
+                    raise safeGraphError(f"""*** [{",".join(errors)}] not available for {product}, use another query""")
             if len(available_columns) > 0:
                 query += pattern_pick["__header__"] + " "
                 for j in available_columns:
@@ -149,7 +157,15 @@ class HTTP_Client:
         elif product == "safegraph_geometry.*":
             # if all data from safegraph_geometry
             pattern_pick = __PATTERNS__["safegraph_geometry"]
-            available_columns = (lambda x: [j for j in pattern_pick if j not in ["__header__", "__footer__"]] if x=="*" else [j for j in pattern_pick if j in columns] )(columns)#[j for j in pattern_pick if j in columns]
+            available_columns = (lambda x: [j for j in pattern_pick if j not in ["__header__", "__footer__"]] if x=="*" else [j for j in pattern_pick if j in cols] )(cols)#[j for j in pattern_pick if j in columns]
+            if type(cols) != str:
+                try:
+                    cols.remove("placekey")
+                except ValueError:
+                    pass
+                if len(available_columns) < len(cols):
+                    errors = [i for i in cols if i not in pattern_pick]
+                    raise safeGraphError(f"""*** [{",".join(errors)}] not available for {product}, use another query""")
             if len(available_columns) > 0:
                 query += pattern_pick["__header__"] + " "
                 for j in available_columns:
@@ -162,7 +178,15 @@ class HTTP_Client:
         elif product == "safegraph_monthly_patterns.*":
             # if all data from safegraph_monthly_patterns
             pattern_pick = __PATTERNS__["safegraph_monthly_patterns"]
-            available_columns = (lambda x: [j for j in pattern_pick if j not in ["__header__", "__footer__"]] if x=="*" else [j for j in pattern_pick if j in columns] )(columns)#[j for j in pattern_pick if j in columns]
+            available_columns = (lambda x: [j for j in pattern_pick if j not in ["__header__", "__footer__"]] if x=="*" else [j for j in pattern_pick if j in cols] )(cols)#[j for j in pattern_pick if j in columns]
+            if type(cols) != str:
+                try:
+                    cols.remove("placekey")
+                except ValueError:
+                    pass
+                if len(available_columns) < len(cols):
+                    errors = [i for i in cols if i not in pattern_pick]
+                    raise safeGraphError(f"""*** [{",".join(errors)}] not available for {product}, use another query""")
             if len(available_columns) > 0:
                 query += pattern_pick["__header__"] + " "
                 for j in available_columns:
@@ -175,7 +199,15 @@ class HTTP_Client:
         elif product == "safegraph_weekly_patterns.*":
             # if all data from safegraph_weekly_patterns
             pattern_pick = __PATTERNS__["safegraph_weekly_patterns"]
-            available_columns = (lambda x: [j for j in pattern_pick if j not in ["__header__", "__footer__"]] if x=="*" else [j for j in pattern_pick if j in columns] )(columns)#[j for j in pattern_pick if j in columns]
+            available_columns = (lambda x: [j for j in pattern_pick if j not in ["__header__", "__footer__"]] if x=="*" else [j for j in pattern_pick if j in cols] )(cols)#[j for j in pattern_pick if j in columns]
+            if type(cols) != str:
+                try:
+                    cols.remove("placekey")
+                except ValueError:
+                    pass
+                if len(available_columns) < len(cols):
+                    errors = [i for i in cols if i not in pattern_pick]
+                    raise safeGraphError(f"""*** [{",".join(errors)}] not available for {product}, use another query""")
             if len(available_columns) > 0:
                 query += pattern_pick["__header__"] + " "
                 for j in available_columns:
@@ -185,9 +217,9 @@ class HTTP_Client:
             else:
                 raise safeGraphError("*** picked column(s) not available")
             data_type = ["safegraph_weekly_patterns"]
-        elif type(columns) != list:
-            raise ValueError("""*** columns argument must to be a list or one of the following string: 
-                * , safegraph_core.* , safegraph_geometry.* , safegraph_monthly_patterns.* , safegraph_weekly_patterns.*
+        else:
+            raise ValueError("""*** product argument must be one of the following string: 
+                core, geometry, monthly_patterns, weekly_patterns
             """)
         return query, data_type
 
@@ -222,8 +254,9 @@ class HTTP_Client:
             for v in self.lst[l].keys():
                 if self.lst[l][v] == None:
                     count+=1
-                if count >= len(self.lst[l])-1:
-                    self.lst.pop()
+                    continue
+            if count >= len(self.lst[l])-1:
+                self.lst.pop(l)
 
     def __lengthCheck__(self, data_frame):
         if len(data_frame) < 1:
@@ -243,12 +276,50 @@ class HTTP_Client:
         for i in range(0, len(lst), n):
             yield lst[i:i + n]
 
-    def _date_setter(self, date):
+    def __date_setter(self, date):
         if date != "__default__":
             self.date = date
         elif date == "__default__":
             self.date = [datetime.now().strftime("%Y-%m-%d")]
 
+    def __prettyfy(self, __query__) -> None:
+        __pick__ = False # picks the next value without additional operation; for first check
+        pre = 0
+        str_ = ""
+        __split__ = re.sub(r'\s+', " ", __query__).split(" ")
+        for s in __split__:
+            pass
+            # v2
+            # if s == "{":
+            #     pre+=1
+            #     str_ += s + "\n" + pre*"\t"
+            # elif s == "}":
+            #     pre-=1
+            #     str_ += s + "\n" + pre*"\t"
+            # else:
+            #     str_+=s
+
+            # v1
+            # if __pick__:
+            #     if s == "{":
+            #         pre+=1
+            #     elif s == "}":
+            #         pre-=1
+            #     str_+= " " + s + "\n" + pre*"\t"
+            #     __pick__ = False
+            # elif s == "{":
+            #     pre+=1
+            #     str_+=" " + s# + "\n" + pre*"\t"
+            # elif s == "}":
+            #     pre-=1
+            #     str_+=" " + s# + "\n" + pre*"\t"
+            # elif s.endswith(":"):
+            #     str_+=" " + s
+            #     __pick__ = True
+            # else:
+            #     str_+= "\n" + pre*"\t" + s
+        print(re.sub(r'\n+', '\n', __query__))
+
     def save(self, path="__default__", return_type="__default__"):
         """
             :param str path:                  (optional) location of the file e.g: "results.csv"
@@ -258,8 +329,8 @@ class HTTP_Client:
             :param str return_type:          (optional) pandas or list 
                 return type of the saved format by default last return format
         """
-        if return_type == "__default__":
-            return_type = self.return_type
+        if return_type != "__default__":
+            self.return_type = return_type
         if self.return_type == "pandas":
             if path != "__default__":
                 self.df.to_csv(path_or_buf=path, index=False)
@@ -267,17 +338,22 @@ class HTTP_Client:
                self.df.to_csv(path_or_buf="results.csv", index=False) 
         elif self.return_type == "list":
             if path != "__default__":
-                with open(path, 'w') as json_file:
-                    json.dump(self.lst, json_file, indent=4)
+                json_file = open(path, 'w')
+                json.dump(self.lst, json_file, indent=4)
+                json_file.close()
             else:
-                with open("results.json", 'w') as json_file:
-                    json.dump(self.lst, json_file, indent=4)
+                json_file = open("results.json", 'w')
+                json.dump(self.lst, json_file, indent=4)
+                json_file.close()
+        else:
+            raise safeGraphError("*** unknown return_type should be a string, either pandas or list")
 
-    def sg_merge(self, datasets:list, how="outer", rt="pd"):
+    def sg_merge(self, datasets:list, how:str="outer", return_type:str="pandas"):
         """
             :param list datasets:           a list of dataframes or jsons
             :param str how:                 (optional) join style either outer or inner       
-            :return:                        The data of given datasets in first index type
+            :param str return_type:         (optional) return type either "pandas" or "list"; default "pandas"       
+            :return:                        The data of given return_type
             :rtype:                         pandas.DataFrame or list
         """
         df = datasets[0]
@@ -295,38 +371,43 @@ class HTTP_Client:
                 except pd.errors.MergeError as e:
                     print(e)
                 except TypeError:
-                    print("*** weekly patterns cannot be merged for the current page: TOFIX")
+                    print("*** weekly_patterns/monthly_patterns cannot be merged for the current page: TOFIX")
             self.__adjustments(df.to_dict("records"))
         elif data_type == list:
-            # change arr's values othervise
+            # change arr's values otherwise
             r_dataset = []
             for i in range(len(datasets)):
                 if type(datasets[i]) != data_type:
                     raise safeGraphError(f"*** each datasets' type must be the same:{data_type}, cannot be {type(i)}")
                 r_dataset.append(pd.DataFrame.from_dict(datasets[i]))
-            return self.sg_merge(r_dataset, how=how, rt="ls")
+            return self.sg_merge(r_dataset, how=how, return_type="list")
 
 
             for i in datasets[1:]:
                 df = df + i
             self.__adjustments(df)  
-        if rt == "pd":      
+        if return_type == "pandas":      
             return df
         else:
             return df.to_dict("records")
 
-    def lookup(self, product:str, placekeys:list, columns, date = '__default__', return_type="pandas"):
+    def lookup(self, product:str, placekeys:list, columns, date='__default__', preview_query:str=False, return_type:str="pandas"):
         """
+            :param str product:             spesific product to pull data from either one core/geometry/weekly_patterns/monthly_patterns
             :param list placekeys:          Unique Placekey ID/IDs inside an array
                 [ a single placekey string or a list of placekeys are both acceptable ]
             :param str return_type:         (optional) pandas or list
                 default -> pandas
             :param columns:                 list or str 
                 "*" as string for all or desired column(s) in a [list]
+            :param date:                    (optional) "YYYY-MM-DD", if not given using today's date
+                list    => ["YYYY-MM-DD", "YYYY-MM-DD", "YYYY-MM-DD"]
+                string  => "YYYY-MM-DD"
+                dict    => {'date_range_start': "YYYY-MM-DD", 'date_range_end': "YYYY-MM-DD"}
             :return:                        The data of given placekeys in return_type
             :rtype:                         pandas.DataFrame or dict
         """
-        self._date_setter(date)
+        self.__date_setter(date)
         self.return_type = return_type
         product = f"safegraph_{product}.*"
         params = {"placekeys": placekeys}
@@ -348,18 +429,18 @@ class HTTP_Client:
             if date == '__default__':
                 dataset = dataset.replace("_DATE_BLOCK_", "")
             else:
-                dataset = dataset.replace("_DATE_BLOCK_", "(date: _DATE_ )")
-                dataset = dataset.replace("_DATE_", f'"{i}"')
+                dataset = dataset.replace("_DATE_BLOCK_", f'(date: "{i}" )')
             # print(dataset+"\n")
-            query = gql(
-                f"""query($placekeys: [Placekey!]) {{
-                    batch_lookup(placekeys: $placekeys) {{
-                        placekey
-                        {dataset}
-                    }}
-                }}"""
-            )
-            
+            __query__ = f"""query($placekeys: [Placekey!]) {{
+ batch_lookup(placekeys: $placekeys) {{
+    placekey
+    {dataset}
+    }}
+}}"""
+            if preview_query:
+                self.__prettyfy(__query__)
+                return
+            query = gql(__query__)            
             result = self.client.execute(query, variable_values=params)
             for place in result['batch_lookup']:
                 dict_ = {}
@@ -372,10 +453,7 @@ class HTTP_Client:
                         pass
                 data_frame.append(dict_)
 
-        # adjustments
-        # self.__lengthCheck__(data_frame) # not working in this function
         self.__adjustments(data_frame)
-
         if self.return_type == "pandas":
             return self.df
         elif self.return_type == "list":
@@ -391,16 +469,22 @@ class HTTP_Client:
             iso_country_code:str=None, 
             postal_code:str=None,
             latitude:float=None, longitude:float=None, 
-            date = '__default__',
-            return_type="pandas"):
+            date='__default__',
+            preview_query:str=False,
+            return_type:str="pandas"):
         """
-            :param str location_name:       location_name of the desidred lookup
-            :param str street_address:      street_address of the desidred lookup
-            :param str city:                city of the desidred lookup
-            :param str region:              region of the desidred lookup
-            :param str iso_country_code:    iso_country_code of the desidred lookup
+            :param str product:             spesific product to pull data from either one core/geometry/weekly_patterns/monthly_patterns
             :param columns:                 list or str 
-                "*" as string for all or desired column(s) in a [list]
+                "*" as string for all or desired column(s) in a list
+            :param date:                    (optional) "YYYY-MM-DD", if not given using today's date
+                list    => ["YYYY-MM-DD", "YYYY-MM-DD", "YYYY-MM-DD"]
+                string  => "YYYY-MM-DD"
+                dict    => {'date_range_start': "YYYY-MM-DD", 'date_range_end': "YYYY-MM-DD"}
+            :param str location_name:       (optional) location_name of the desidred lookup
+            :param str street_address:      (optional) street_address of the desidred lookup
+            :param str city:                (optional) city of the desidred lookup
+            :param str region:              (optional) region of the desidred lookup
+            :param str iso_country_code:    (optional) iso_country_code of the desidred lookup
             :param str return_type:         (optional) pandas or list
                 default -> pandas
             :return:                        The data of given placekey in return_type
@@ -422,7 +506,7 @@ When querying by location & address, it's necessary to have at least the followi
     location_name + street_address + postal_code + iso_country_code
     location_name + latitude + longitude + iso_country_code
             ''')
-        self._date_setter(date)
+        self.__date_setter(date)
         self.return_type = return_type
         product = f"safegraph_{product}.*"
         params = f"""
@@ -449,18 +533,22 @@ When querying by location & address, it's necessary to have at least the followi
                 dataset, data_type = self.__dataset_WM(product, columns)
                 if dataset == "":
                     continue
-            dataset = dataset.replace("_DATE_", f'"{i}"')
+            if date == '__default__':
+                dataset = dataset.replace("_DATE_BLOCK_", "")
+            else:
+                dataset = dataset.replace("_DATE_BLOCK_", f'(date: "{i}" )')
             # print(dataset+"\n")
-            query = gql(
-                f"""query {{
-                    lookup(query: {{
-                        {params}
-                        }}) {{ 
-                        placekey 
-                        {dataset}
-                    }}
-                }}"""
-            )
+            __query__ = f"""query {{
+ lookup(query: {{{params}
+ }}) {{ 
+        placekey 
+        {dataset}
+    }}
+}}"""
+            if preview_query:
+                self.__prettyfy(__query__)  
+                return
+            query = gql(__query__)
             result = self.client.execute(query)
             dict_ = {}
             for j in data_type:
@@ -475,10 +563,7 @@ When querying by location & address, it's necessary to have at least the followi
                 raise safeGraphError("*** query didnt match with an item")
             data_frame.append(dict_)
 
-        # adjustments
-        # self.__lengthCheck__(data_frame) # not working in this function
         self.__adjustments(data_frame)
-
         if self.return_type == "pandas":
             return self.df
         elif self.return_type == "list":
@@ -486,22 +571,28 @@ When querying by location & address, it's necessary to have at least the followi
         else:
             raise safeGraphError(f'return_type "{return_type}" does not exist')
 
-    def search(self, product, columns,
+    def search(self, product, columns, date='__default__',
         # params
         brand:str=None, brand_id:str=None, naics_code:int=None, phone_number:str=None,
         # address with following sub-fields
         location_name:str=None, street_address:str=None, city:str=None, region:str=None, postal_code:str=None, iso_country_code:str=None,
         max_results:int=20,
         after_result_number:int=0,
-        date:str='__default__',
+        preview_query:str=False,
         return_type:str="pandas"):
         """
+            :param str product:             spesific product to pull data from either one core/geometry/weekly_patterns/monthly_patterns
             :param columns:                 list or str 
-                "*" as string for all or desired column(s) in a [list]
+                "*" as string for all or desired column(s) in a list
+            :param date:                    (optional) "YYYY-MM-DD", if not given using today's date
+                list    => ["YYYY-MM-DD", "YYYY-MM-DD", "YYYY-MM-DD"]
+                string  => "YYYY-MM-DD"
+                dict    => {'date_range_start': "YYYY-MM-DD", 'date_range_end': "YYYY-MM-DD"}
             :param str brand:               brand for searching query
             :param str brand_id:            brand_id for searching query
             :param int naics_code:          naics_code for searching query
             :param str phone_number:        phone_number for searching query
+            :param str location_name:       location_name for searching query
             :param str street_address:      street_address of the desidred place
             :param str city:                city of the desidred place
             :param str region:              region of the desidred place
@@ -513,12 +604,13 @@ When querying by location & address, it's necessary to have at least the followi
                 default -> pandas
             :return:                        The data of given placekey in return_type
             :rtype:                         pandas.DataFrame or dict
-        """                               ############ 
-        #################################################        |```|  /\   |````|
-        self.max_results = max_results    ##################     |\``  / _\  |    |
-        self.return_type = return_type    ###################    | \  /    \ |____|__
-        self._date_setter(date) ###########
-        #################################################
+        """
+                                          ############
+        #################################################        |```|  /\   |`````|
+        self.max_results = max_results    ##################     |   | /  \  |     |
+        self.return_type = return_type    ###################    |`\` /____\ |     |
+        self.__date_setter(date)           #################      |  \/      \|     |
+        #################################################        |  /\       |\____|__
                                           ############
         product = f"safegraph_{product}.*"
         params = f"""
@@ -534,7 +626,7 @@ When querying by location & address, it's necessary to have at least the followi
 {(lambda x,y: f' {x}: "{y}" ' if y!=None else "")("region", region)}
 {(lambda x,y: f' {x}: "{y}" ' if y!=None else "")("postal_code", postal_code)}
 {(lambda x,y: f' {x}: "{y}" ' if y!=None else "")("iso_country_code", iso_country_code)}
-}}"""
+ }}"""
         if address != ' address: {\n\n\n\n\n\n\n}':
             params+=address
         after = 0
@@ -557,24 +649,27 @@ When querying by location & address, it's necessary to have at least the followi
                     dataset, data_type = self.__dataset_WM(product, columns)
                     if dataset == "":
                         continue
-                dataset = dataset.replace("_DATE_", f'"{i}"')
+                if date == '__default__':
+                    dataset = dataset.replace("_DATE_BLOCK_", "")
+                else:
+                    dataset = dataset.replace("_DATE_BLOCK_", f'(date: "{i}" )')
                 # print(dataset+"\n")
-                query = gql(
-                    f"""query {{
-                        search(first: {first} after: {after+after_result_number} filter: {{
-                            {params}
-                            }}) {{ 
-                            placekey 
-                            {dataset}
-                        }}
-                    }}"""
-                )
+                __query__ = f"""query {{
+ search(first: {first} after: {after+after_result_number} filter: {{{params}   
+ }}) {{ 
+        placekey 
+        {dataset}
+    }}
+}}"""
+                if preview_query:
+                    self.__prettyfy(__query__) 
+                    return
+                query = gql(__query__)
                 try:
                     result = self.client.execute(query)
                 except Exception as e:
                     print("\n\n\n\t*** ERROR ***\n\n\n")
                     raise e
-                    #if type(e) == ConnectionError:
                     result = {'search': []}
                     self.errors.append(f"{after+after_result_number}-{first+after+after_result_number}")
                 output+=result['search']
@@ -586,11 +681,8 @@ When querying by location & address, it's necessary to have at least the followi
                         dict_.update(out[j])
                         dict_['placekey'] = (out['placekey'])
                     except TypeError:
-                        # 'safegraph_weekly_patterns': None
                         pass
-                # dict_['placekey'] = out["placekey"]
                 data_frame.append(dict_)
-
         self.__lengthCheck__(data_frame)
         self.__adjustments(data_frame)
         self.__error_check(after_result_number)
